@@ -258,7 +258,7 @@ class WeakrefableConnection(sqlite3.Connection):
         dbpp = get_dbpp(self)
 
         # get db pointer
-        db_ptr = dbpp.contents
+        db_ptr = dbpp.contents.value
 
         # we need to call call base method
         # to finalize all statements before closing database
@@ -272,7 +272,7 @@ class WeakrefableConnection(sqlite3.Connection):
             raise
 
         if db_ptr:
-            rc = sqlite3_close(dbptr)
+            rc = sqlite3_close(db_ptr)
             if rc:
                 raise Exception("Error while creating function: %r" % (rc,))
 
@@ -606,9 +606,6 @@ class StreamingFile(object):
 
                     self._start_thread()
 
-                    # in case we loop again below
-                    ctx = self.reset_lock.shared_context()
-
                 if not size:
                     return b''
 
@@ -626,6 +623,9 @@ class StreamingFile(object):
 
                 redo = self._wait_for_range(offset, size)
                 if redo:
+                    # generator context managers are one-time use only
+                    # so recreate next one
+                    ctx = self.reset_lock.shared_context()
                     continue
 
                 # TODO: port to windows, can use ReadFile
@@ -1031,6 +1031,7 @@ class CachedFile(object):
                 finally:
                     towrite.close()
 
+                new_stat = None
                 if md is not None:
                     assert md.id == self._id, \
                         "Bad assumption on how overwrite works :("
@@ -1074,7 +1075,8 @@ class CachedFile(object):
                     self._complete_tag = sync_tag
                     self._upload_now = None
                     self._upload_cond.notify_all()
-                    self._file.replace_underlying(StreamingFile(self._fs, new_stat))
+                    if new_stat is not None:
+                        self._file.replace_underlying(StreamingFile(self._fs, new_stat))
 
                 self._fs._submit_write(self._id, md)
             except Exception:
