@@ -25,10 +25,10 @@ import io
 import json
 import logging
 import os
-import random
 import subprocess
 import sys
 import syslog
+import time
 import urllib.request
 
 import appdirs
@@ -44,7 +44,7 @@ from keyring.errors import KeyringError
 
 import sentry_sdk
 
-from block_tracing import block_tracing, BLOCK_TRACING_INHERITS
+from block_tracing import block_tracing
 
 from dbxfs.dbxfs import FileSystem as DropboxFileSystem
 from dbxfs.cachingfs import FileSystem as CachingFileSystem, check_runtime_requirements
@@ -63,7 +63,6 @@ APP_NAME = "dbxfs"
 
 # This exposure is intentional
 APP_KEY = "iftkeq2y4qj0nbt"
-APP_SECRET = "y245xn4rg4lf0it"
 
 def yes_no_input(message=None, default_yes=False):
     if default_yes:
@@ -81,8 +80,8 @@ def yes_no_input(message=None, default_yes=False):
 def parse_encrypted_folder_arg(string):
     return dict(path=string)
 
-def base_create_fs(access_token, cache_folder):
-    fs = CachingFileSystem(DropboxFileSystem(access_token), cache_folder=cache_folder)
+def base_create_fs(client_kw_args, cache_folder):
+    fs = CachingFileSystem(DropboxFileSystem(client_kw_args), cache_folder=cache_folder)
 
     # From a purity standpoint the following layer ideally would
     # go between the caching fs and dropbox fs, but because the
@@ -116,8 +115,7 @@ class RealSysLogHandler(logging.Handler):
 
 def on_new_process(proc_args):
     # Protect access token and potentially encryption keys
-    if not BLOCK_TRACING_INHERITS:
-        block_tracing()
+    block_tracing()
 
     display_name = proc_args['display_name']
     verbose = int(proc_args['verbose'])
@@ -128,11 +126,27 @@ def on_new_process(proc_args):
     level = [logging.WARNING, logging.INFO, logging.DEBUG][min(2, verbose)]
     logging.basicConfig(level=level, handlers=[logging_stream], format=format_)
 
+    if int(proc_args.get('send_error_reports', '0')):
+        version = proc_args['version']
+        try:
+            sentry_sdk.init("https://b4b13ebd300849bd92260507a594e618@sentry.io/1293235",
+                            release='%s@%s' % (APP_NAME, version),
+                            with_locals=False)
+        except Exception:
+            log.warning("Failed to initialize sentry", exc_info=True)
+
 def create_fs(fs_args):
-    access_token = fs_args['access_token']
+    refresh_token = fs_args.get('refresh_token')
+    access_token = fs_args.get('access_token')
     cache_folder = fs_args['cache_folder']
 
-    sub_create_fs = functools.partial(base_create_fs, access_token, cache_folder)
+    dbx_client_args = dict(
+        oauth2_refresh_token=refresh_token,
+        oauth2_access_token=access_token,
+        app_key=APP_KEY,
+    )
+
+    sub_create_fs = functools.partial(base_create_fs, dbx_client_args, cache_folder)
 
     if safefs_wrap_create_fs is not None:
         sub_create_fs = safefs_wrap_create_fs(sub_create_fs, fs_args)
@@ -224,6 +238,8 @@ def _main(argv=None):
                         help="print default config file path to standard out and quit")
     parser.add_argument("--cache-dir",
                         help="file cache directory")
+    parser.add_argument("--get-refresh-token", action='store_true',
+                        help="authorize new refresh token and print to standard out")
     parser.add_argument("mount_point", nargs='?')
     args = parser.parse_args(argv[1:])
 
@@ -260,104 +276,140 @@ def _main(argv=None):
         print(config_file)
         return 0
 
-    try:
-        os.makedirs(config_dir, exist_ok=True)
-    except OSError as e:
-        print("Unable to create configuration directory: %s" % (e,))
-        return -1
-
     config = {}
-    try:
-        f = open(config_file)
-    except IOError as e:
-        if e.errno != errno.ENOENT: raise
-    else:
-        try:
-            with f:
-                config = json.load(f)
-        except ValueError as e:
-            print("Config file %r is not valid json: %s" % (config_file, e))
-            return -1
-
-    cache_folder = args.cache_dir
-    if cache_folder is None:
-        cache_folder = config.get("cache_dir")
-
-    mount_point = args.mount_point
-    if mount_point is None:
-        mount_point = config.get("mount_point")
-
-    if not args.smb_no_mount and mount_point is None:
-        parser.print_usage()
-        print("%s: error: please provide the mount_point argument" % (os.path.basename(argv[0]),))
-        return 1
-
-    encrypted_folders = config.get("encrypted_folders", []) + args.encrypted_folders
 
+    refresh_token = None
     access_token = None
-    save_access_token = False
+    save_refresh_token = False
     save_config = False
+    came_from_command = False
+    came_from_keyring = False
 
-    access_token_command = config.get("access_token_command", None)
-    if access_token_command is not None:
-        print("Running %r for access token" % (' '.join(access_token_command),))
+    if not args.get_refresh_token:
         try:
-            access_token = subprocess.check_output(access_token_command).decode("utf-8")
-        except UnicodeDecodeError:
-            print("Access token command output is not utf-8 encoded")
+            os.makedirs(config_dir, exist_ok=True)
+        except OSError as e:
+            print("Unable to create configuration directory: %s" % (e,))
             return -1
-        except TypeError:
-            print("Bad access token command: %r, " % (access_token_command,))
-            return -1
-        # NB: access tokens never contain white-space and the access token
-        #     command often accidentally appends a newline character.
-        access_token = access_token.strip()
 
-    if access_token is None:
-        keyring_user = config.get("keyring_user", None)
-
-        if keyring_user is not None:
+        try:
+            f = open(config_file)
+        except IOError as e:
+            if e.errno != errno.ENOENT: raise
+        else:
             try:
-                access_token = keyring.get_password(APP_NAME, keyring_user)
-            except KeyringError as e:
-                print("Failed to get access token from keyring: %s" % (e,))
+                with f:
+                    config = json.load(f)
+            except ValueError as e:
+                print("Config file %r is not valid json: %s" % (config_file, e))
+                return -1
 
-    if access_token is None:
-        access_token_privy = config.get("access_token_privy", None)
-        if access_token_privy is not None:
-            passwd = None
-            while True:
-                passwd = getpass.getpass("Enter access token passphrase (not your Dropbox password) (Ctrl-C to quit): ")
+        cache_folder = args.cache_dir
+        if cache_folder is None:
+            cache_folder = config.get("cache_dir")
+
+        mount_point = args.mount_point
+        if mount_point is None:
+            mount_point = config.get("mount_point")
+
+        if not args.smb_no_mount and mount_point is None:
+            parser.print_usage()
+            print("%s: error: please provide the mount_point argument" % (os.path.basename(argv[0]),))
+            return 1
+
+        encrypted_folders = config.get("encrypted_folders", []) + args.encrypted_folders
+
+        if refresh_token is None and access_token is None:
+            refresh_token_command = config.get("refresh_token_command", None)
+            if refresh_token_command is not None:
+                print("Running %r for refresh token" % (' '.join(refresh_token_command),))
                 try:
-                    access_token = privy.peek(access_token_privy, passwd).decode('utf-8')
-                except ValueError:
-                    if not yes_no_input("Incorrect password, create new access token?"):
-                        continue
-                break
-            del passwd
+                    refresh_token = subprocess.check_output(refresh_token_command).decode("utf-8")
+                except UnicodeDecodeError:
+                    print("Refresh token command output is not utf-8 encoded")
+                    return -1
+                except TypeError:
+                    print("Bad refresh token command: %r, " % (refresh_token_command,))
+                    return -1
+                # NB: refresh tokens never contain white-space and the refresh token
+                #     command often accidentally appends a newline character.
+                refresh_token = refresh_token.strip()
+                came_from_command = True
+
+        access_token_command = config.get("access_token_command", None)
+        if refresh_token is None and access_token is None and access_token_command is not None:
+            print("Running %r for access token" % (' '.join(access_token_command),))
+            try:
+                access_token = subprocess.check_output(access_token_command).decode("utf-8")
+            except UnicodeDecodeError:
+                print("Access token command output is not utf-8 encoded")
+                return -1
+            except TypeError:
+                print("Bad access token command: %r, " % (access_token_command,))
+                return -1
+            # NB: access tokens never contain white-space and the access token
+            #     command often accidentally appends a newline character.
+            access_token = access_token.strip()
+            came_from_command = True
+
+        if refresh_token is None and access_token is None:
+            refresh_token_privy = config.get("refresh_token_privy", None)
+            if refresh_token_privy is not None:
+                if not sys.stdout.isatty():
+                    print("Error: need to run in interactive mode to get password to decrypt refresh token")
+                    return 1
+
+                passwd = None
+                while True:
+                    passwd = getpass.getpass("Enter refresh token passphrase (not your Dropbox password) (Ctrl-C to quit): ")
+                    try:
+                        refresh_token = privy.peek(refresh_token_privy, passwd).decode('utf-8')
+                    except ValueError:
+                        if not yes_no_input("Incorrect password, create new refresh token?"):
+                            continue
+                    break
+                del passwd
+
+        if refresh_token is None and access_token is None:
+            keyring_user = config.get("keyring_user", None)
+
+            if keyring_user is not None:
+                try:
+                    access_token = keyring.get_password(APP_NAME, keyring_user)
+                except KeyringError as e:
+                    print("Failed to get access token from keyring: %s" % (e,))
+                else:
+                    came_from_keyring = True
+
+        if refresh_token is None and access_token is None:
+            access_token_privy = config.get("access_token_privy", None)
+            if access_token_privy is not None:
+                if not sys.stdout.isatty():
+                    print("Error: need to run in interactive mode to get password to decrypt access token")
+                    return 1
+
+                passwd = None
+                while True:
+                    passwd = getpass.getpass("Enter access token passphrase (not your Dropbox password) (Ctrl-C to quit): ")
+                    try:
+                        access_token = privy.peek(access_token_privy, passwd).decode('utf-8')
+                    except ValueError:
+                        if not yes_no_input("Incorrect password, create new access token?"):
+                            continue
+                    break
+                del passwd
 
-    try_directly = False
     while True:
-        if access_token is None:
-            save_access_token = True
+        if access_token is None and refresh_token is None:
+            if not sys.stdout.isatty():
+                print("Error: run in an interactive shell to obtain a new refresh token")
+                return 1
 
-        if (access_token is None and
-            try_directly and
-            yes_no_input("Want to try entering the access token directly?")):
-            print("Go to https://dropbox.com/developers/apps to "
-                  "create an app and generate a personal access token.")
+            save_refresh_token = True
 
-            while True:
-                access_token = getpass.getpass("Enter Access token (Ctrl-C to quit): ")
-                if not access_token:
-                    print("Access tokens cannot be empty")
-                    continue
-                break
-
-        if access_token is None:
-            auth_flow = dropbox.DropboxOAuth2FlowNoRedirect(APP_KEY, APP_SECRET)
+            auth_flow = dropbox.DropboxOAuth2FlowNoRedirect(APP_KEY, use_pkce=True, token_access_type='offline')
             authorize_url = auth_flow.start()
-            print("We need an access token. Perform the following steps:")
+            print("We need a refresh token. Perform the following steps:")
             print("1. Go to " + authorize_url)
             print("2. Click \"Allow\" (you may have to log in first)")
             print("3. Copy the authorization code.")
@@ -373,33 +425,60 @@ def _main(argv=None):
                 oauth_result = auth_flow.finish(auth_code)
             except Exception as e:
                 print("Authorization code was invalid!")
-                try_directly = True
                 continue
 
-            access_token = oauth_result.access_token
+            refresh_token = oauth_result.refresh_token
 
         # test out access token
         try:
-            dropbox.Dropbox(access_token).users_get_current_account()
+            assert (refresh_token is None) != (access_token is None)
+            dropbox.Dropbox(
+                oauth2_access_token=access_token,
+                oauth2_refresh_token=refresh_token,
+                app_key=APP_KEY,
+            ).users_get_current_account()
         except (dropbox.exceptions.BadInputError,
                 dropbox.exceptions.AuthError,
                 ValueError) as e:
-            print("Error using access token: %s" % (e,))
-            access_token = None
-            try_directly = True
-        except OSError:
-            if not yes_no_input("Error connecting to Dropbox, Try again?"):
+            if came_from_command:
+                if access_token is not None:
+                    print("""\
+The access token retrieved by executing the command specified in the
+'access_token_command' configuration key is no longer valid, please
+obtain a new one.\
+""")
+                else:
+                    print("""\
+The refresh token retrieved by executing the command specified in the
+'refresh_token_command' configuration key is no longer valid, please
+obtain a new one.\
+""")
                 return 1
+            print("Token was invalid: %s" % (e,))
+            refresh_token = None
+            access_token = None
+            came_from_keyring = False
+        except OSError:
+            if sys.stdout.isatty():
+                if not yes_no_input("Error connecting to Dropbox, Try again?"):
+                    return 1
+            else:
+                time.sleep(1)
         else:
             break
 
-    if save_access_token and yes_no_input("We're all connected. Do you want to save your credentials for future runs?", default_yes=True):
-        keyring_user = ''.join([random.choice("asdfghjklzxcvbnmqwertyuiop")
-                                for _ in range(24)])
-        try:
-            keyring.set_password(APP_NAME, keyring_user, access_token)
-        except (KeyringError, RuntimeError) as e:
-            print("We need a passphrase to encrypt your access token before we can save it.")
+    if args.get_refresh_token:
+        print("Refresh token:", refresh_token)
+        return 0
+
+    if came_from_keyring:
+        assert access_token is not None
+
+        print("""\
+Warning: the use of the system keyring to store the access token is
+deprecated, this functionality will be removed in the future.
+""")
+        if sys.stdout.isatty() and yes_no_input("Would you like to store the access token in a password-encrypted file instead?"):
             print("Warning: Your access token passphrase must contain enough randomness to be resistent to hacking. You can read this for more info: https://blogs.dropbox.com/tech/2012/04/zxcvbn-realistic-password-strength-estimation/")
             while True:
                 pass_ = getpass.getpass("Enter new access token passphrase: ")
@@ -409,22 +488,44 @@ def _main(argv=None):
                 else:
                     del pass2_
                     break
-            config.pop('keyring_user', None)
+            try:
+                keyring.delete_password(APP_NAME, config['keyring_user'])
+            except Exception:
+                log.warning("Failed to delete token from keyring, it must be done manually", exc_info=True)
             config['access_token_privy'] = privy.hide(access_token.encode('utf-8'), pass_, server=False)
+            config.pop('keyring_user', None)
             del pass_
             save_config = True
-        else:
-            config.pop('access_token_privy', None)
-            config['keyring_user'] = keyring_user
-            save_config = True
 
-    if not config.get("asked_send_error_reports", False):
+    if save_refresh_token and yes_no_input("We're all connected. Do you want to save your credentials for future runs?", default_yes=True):
+        assert refresh_token is not None
+
+        # clear all token commands
+        for key in ['access_token_privy', "keyring_user",
+                    'refresh_token_privy']:
+            config.pop(key, None)
+
+        print("We need a passphrase to encrypt your refresh token before we can save it.")
+        print("Warning: Your refresh token passphrase must contain enough randomness to be resistent to hacking. You can read this for more info: https://blogs.dropbox.com/tech/2012/04/zxcvbn-realistic-password-strength-estimation/")
+        while True:
+            pass_ = getpass.getpass("Enter new refresh token passphrase: ")
+            pass2_ = getpass.getpass("Enter new refresh token passphrase (again): ")
+            if pass_ != pass2_:
+                print("Passphrases didn't match, please re-enter")
+            else:
+                del pass2_
+                break
+        config['refresh_token_privy'] = privy.hide(refresh_token.encode('utf-8'), pass_, server=False)
+        del pass_
+        save_config = True
+
+    if sys.stdout.isatty() and not config.get("asked_send_error_reports", False):
         if yes_no_input("Would you like to help us improve %s by providing anonymous error reports?" % (APP_NAME,), default_yes=True):
             config['send_error_reports'] = True
         config['asked_send_error_reports'] = True
         save_config = True
 
-    if save_access_token and yes_no_input("Do you want \"%s\" to be the default mount point?" % (mount_point,), default_yes=True):
+    if save_refresh_token and yes_no_input("Do you want \"%s\" to be the default mount point?" % (mount_point,), default_yes=True):
         config['mount_point'] = mount_point
         save_config = True
 
@@ -456,20 +557,31 @@ def _main(argv=None):
             return 1
         log.debug("Using custom cache path %s", cache_folder)
 
+    assert (refresh_token is None) != (access_token is None)
+
     fs_args = {}
 
-    fs_args['access_token'] = access_token
+    if access_token is not None:
+        fs_args['access_token'] = access_token
+    if refresh_token is not None:
+        fs_args['refresh_token'] = refresh_token
     fs_args['cache_folder'] = cache_folder
 
     if safefs_add_fs_args is not None:
-        with contextlib.closing(base_create_fs(access_token, cache_folder)) as fs:
+        dbx_client_args = dict(
+            oauth2_refresh_token=refresh_token,
+            oauth2_access_token=access_token,
+            app_key=APP_KEY,
+        )
+
+        with contextlib.closing(base_create_fs(dbx_client_args, cache_folder)) as fs:
             safefs_add_fs_args(fs, encrypted_folders, fs_args)
     elif encrypted_folders:
         print("safefs not installed, can't transparently decrypt encrypted folders")
         return 1
 
     if mount_point is not None and not os.path.exists(mount_point):
-        if yes_no_input("Mount point \"%s\" doesn't exist, do you want to create it?" % (mount_point,), default_yes=True):
+        if sys.stdout.isatty() and yes_no_input("Mount point \"%s\" doesn't exist, do you want to create it?" % (mount_point,), default_yes=True):
             try:
                 os.makedirs(mount_point, exist_ok=True)
             except OSError as e:
@@ -481,6 +593,8 @@ def _main(argv=None):
     proc_args = {}
     proc_args['display_name'] = display_name
     proc_args['verbose'] = str(args.verbose)
+    proc_args['version'] = version
+    proc_args['send_error_reports'] = str(int(config.get('send_error_reports', False)))
 
     return userspacefs.simple_main(mount_point, display_name,
                                    ('dbxfs.main.create_fs', fs_args),
